;; -*- geiser-scheme-implementation: guile -*-

;; enemy.scm - This file is part of Guile TD
;; Copyright (C) 2017  Christopher Howard

;; Guile TD is free software: you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; Guile TD is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

(define-module (td enemy)
  #:use-module (td fset)
  #:use-module (td util)
  #:use-module (td tick)
  #:use-module (td hooks)
  #:use-module (td euclidean)
  #:use-module (td picture)
  #:use-module (td gfx)
  #:use-module (aa-tree)
  #:use-module (chickadee)
  #:use-module (chickadee math vector)
  #:use-module (chickadee scripting)
  #:use-module (srfi srfi-1)
  #:use-module (oop goops)
  #:export (affect-speed!
            damage!
            enemies
            <enemy>
            health-bar
            march-enemies!
            <marching>
            launch-destroyer!
            launch-helo!
            launch-trike!
            launch-ptboat!
            launch-soldier!
            launch-soldier-l2!
            launch-soldier-l3!
            launch-tank!
            launch-tank-l2!
            launch-tank-l3!
            launch-tracked-soldier!
            .prev-x
            .prev-y
            slow))

(define enemies #nil)

(add-hook! new-game-hook
           (lambda ()
             (set! enemies #nil)))

(define-class <enemy> (<euclidean>)
  (prev-x #:accessor .prev-x #:init-keyword #:prev-x)
  (prev-y #:accessor .prev-y #:init-keyword #:prev-y)
  (facing #:init-value #:west #:accessor .facing #:init-keyword #:facing)
  (waypoints #:init-keyword #:waypoints #:accessor .waypoints)
  (health #:init-value #nil #:init-keyword #:health)
  (health-tenth #:init-value #nil)
  (speed-λ #:init-value (lambda (s) (base-speed s)) #:accessor speed-λ))

(define-method (.health (e <enemy>) x)
  (slot-set! e 'health x))

(define-method (.health (e <enemy>))
  (let ((slot-val (slot-ref e 'health)))
    (if (nil? slot-val)
        (begin (.health e (.max-health e))
               (.max-health e))
        slot-val)))

(define-method (.health-tenth (e <enemy>) x)
  (slot-set! e 'health-tenth x))

(define-method (.health-tenth (e <enemy>))
  (let ((slot-val (slot-ref e 'health-tenth)))
    (if (nil? slot-val)
        (begin (.health-tenth e (calculate-tenth e))
               (.health-tenth e))
        slot-val)))

(define-method (calculate-tenth (e <enemy>))
  (quotient (round (* 10 (.health e)))
            (round (.max-health e))))
                                
(define-method (< (l1 <list>) (l2 <list>))
  (cond ((and (not (nil? l1)) (nil? l2)) #f)
        ((and (nil? l1) (nil? l2)) #f)
        ((and (nil? l1) (not (nil? l2))) #t)
        ((< (car l1) (car l2)) #t)
        ((equal? (car l1) (car l2)) (< (cdr l1) (cdr l2)))
        (else #f)))

(define (damage! enemy-uid damage-points)
  (catch 'key-not-found
    (lambda ()
      (let* ((enemy (aa-search enemies enemy-uid))
             (new-hp (- (.health enemy) damage-points)))
        (if (<= new-hp 0)
            (set! enemies (aa-delete enemies enemy-uid))
            (let* ((new-enemy (slot-fset enemy 'health new-hp))
                   (tenth-e (.health-tenth new-enemy (calculate-tenth new-enemy))))
              (set! enemies (aa-insert enemies enemy-uid new-enemy))))))
    (lambda (_ _') #nil)))

(define (simple-march m ms speed)
  (let* ((max-pixels (* ms speed))
         (next-waypoint (if (null? (.waypoints m))
                            (throw 'arrived)
                            (first (.waypoints m))))
         (wp-x (first next-waypoint))
         (wp-y (second next-waypoint))
         (m-x (.x m))
         (m-y (.y m)))
          ;; (if (< (.slow-time-left m) 1)
          ;;     (.speed m)
          ;;     (* (.speed m) 0.5))))
    (if (and (= wp-x m-x) (= wp-y m-y))
        (march (slot-fset m 'waypoints (cdr (.waypoints m))) ms)
        (let* ((d-north (max (- wp-y m-y) 0))
               (d-south (max (- m-y wp-y) 0))
               (d-east (max (- wp-x m-x) 0))
               (d-west (max (- m-x wp-x) 0))
               (to-move-v (if (> d-north d-south)
                              (min d-north (* speed ms))
                              (* -1 (min d-south (* speed ms)))))
               (to-move-h (if (> d-east d-west)
                              (min d-east (* speed ms))
                              (* -1 (min d-west (* speed ms))))))
          (clone m
                 ((.prev-x) m-x)
                 ((.prev-y) m-y)
                 ((.x) (+ to-move-h m-x))
                 ((.y) (+ to-move-v m-y))
                 ((.facing)
                  (cond
                   ;; top-right
                   ((and (>= to-move-v 0) (>= to-move-h 0))
                    (if (< to-move-v to-move-h)
                        #:east
                        #:north))
                   ;; bottom-right
                   ((and (< to-move-v 0) (>= to-move-h 0))
                    (if (< (abs to-move-v) to-move-h)
                        #:east
                        #:south))
                   ;; top-left
                   ((and (>= to-move-v 0) (< to-move-h 0))
                    (if (< to-move-v (abs to-move-h))
                        #:west
                        #:north))
                   ;; bottom-left
                   (else
                    (if (< (abs to-move-v) (abs to-move-h))
                        #:west
                        #:south)))))))))

(define (march-enemies!)
  (set! enemies
    (let ((new-tree #nil))
      (aa-for-each (lambda (enemy-pair)
                     (let ((uid (car enemy-pair))
                           (enemy (cdr enemy-pair)))
                       (catch 'arrived
                         (lambda () (set! new-tree
                                      (aa-insert new-tree
                                                 uid
                                                 (march enemy period))))
                         (lambda _
                           (run-hook field-crossed)
                           #nil))))
                   enemies)
      new-tree)))

(add-hook! update-hook (lambda (dt)
                         (unless (game-paused?)
                           (march-enemies!))))

(define-method (health-bar (e <enemy>))
  ;; OPTIMIZE: health-var could be stored in enemy and adjusted at
  ;; health change
  (make <line>
    #:start (vec2 -5 20)
    #:end (vec2 (+ -5 (.health-tenth e)) 20)))

(define-method (march (e <enemy>) ms)
  (simple-march e ms (speed e)))

(define-method (speed (e <enemy>))
  ((speed-λ e) e))

(define-method (affect-speed! uid (e <enemy>) factor ms)
  (let* ((orig-λ (speed-λ e))
         (ending-tick (+ (agenda-time) ms))
         (new-λ (lambda (_)
                  (if (< (agenda-time) ending-tick)
                      (min (orig-λ e) (* (base-speed e) factor))
                      (begin
                        (set! enemies (aa-insert enemies
                                                 uid
                                                 (clone e ((speed-λ) orig-λ))))
                        (orig-λ e))))))
    (set! enemies (aa-insert enemies
                             uid
                             (clone e ((speed-λ) new-λ))))))

(define-class <soldier> (<enemy>)
  (base-speed #:init-value 0.02 #:accessor base-speed)
  (max-health #:init-value 6 #:accessor .max-health)
  (image-west #:init-value #:soldier-west #:accessor image-west)
  (image-north #:init-value #:soldier-north #:accessor image-north)
  (image-south #:init-value #:soldier-south #:accessor image-south)
  (image-east #:init-value #:soldier-east #:accessor image-east)
  (step-seed #:init-thunk (lambda ()
                            (random 1167))
             #:accessor step-seed))

(define-method (draw (si <soldier>) position)
  (let ((r (remainder (agenda-time) (+ (step-seed si) 1167)))
        (basic-sprite
         (get-sprite
          (case (.facing si)
            ((#:west) (image-west si))
            ((#:north) (image-north si))
            ((#:south) (image-south si))
            ((#:east) (image-east si))))))
    (draw (transform basic-sprite
                 (cond ((< r 333)
                        (case (.facing si)
                          ((#:west) (vec2 -3 0))
                          ((#:north) (vec2 0 3))
                          ((#:south) (vec2 0 -3))
                          ((#:east) (vec2 3 0))))
                       ((< r 666)
                        (vec2 0 0))
                       ((< r 999)
                        (case (.facing si)
                          ((#:west) (vec2 3 0))
                          ((#:north) (vec2 0 -3))
                          ((#:south) (vec2 0 3))
                          ((#:east) (vec2 -3 0))))
                       (else
                        (vec2 0 0)))) position)))

(define (launch-soldier! path)
  (set! enemies
    (aa-insert enemies
               (get-uid)
               (make <soldier>
                 #:x (first (first path)) #:y (second (first path))
                 #:prev-x (first (first path)) #:prev-y (second (first path))
                 #:waypoints path))))

(define-class <soldier-l2> (<soldier>)
  (max-health #:init-value 30 #:accessor .max-health)
  (image-west #:init-value #:soldier-l2-west #:accessor image-west)
  (image-north #:init-value #:soldier-l2-north #:accessor image-north)
  (image-south #:init-value #:soldier-l2-south #:accessor image-south)
  (image-east #:init-value #:soldier-l2-east #:accessor image-east)
  (step-seed #:init-thunk (lambda ()
                            (random 1167))
             #:accessor step-seed))

(define (launch-soldier-l2! path)
  (set! enemies
    (aa-insert enemies
               (get-uid)
               (make <soldier-l2>
                 #:x (first (first path)) #:y (second (first path))
                 #:prev-x (first (first path)) #:prev-y (second (first path))
                 #:waypoints path))))

(define-class <soldier-l3> (<soldier-l2>)
  (max-health #:init-value 60 #:accessor .max-health)
  (image-west #:init-value #:soldier-l3-west #:accessor image-west)
  (image-north #:init-value #:soldier-l3-north #:accessor image-north)
  (image-south #:init-value #:soldier-l3-south #:accessor image-south)
  (image-east #:init-value #:soldier-l3-east #:accessor image-east)
  (step-seed #:init-thunk (lambda ()
                            (random 1167))
             #:accessor step-seed))

(define (launch-soldier-l3! path)
  (set! enemies
    (aa-insert enemies
               (get-uid)
               (make <soldier-l3>
                 #:x (first (first path)) #:y (second (first path))
                 #:prev-x (first (first path)) #:prev-y (second (first path))
                 #:waypoints path))))

(define-class <trike> (<enemy>)
  (base-speed #:init-value 0.1 #:accessor base-speed)
  (max-health #:init-value 8 #:accessor .max-health))

(define-method (draw (e <trike>) position)
  (draw (get-sprite (case (.facing e)
                      ((#:west) #:trike-west)
                      ((#:north) #:trike-north)
                      ((#:south) #:trike-south)
                      ((#:east) #:trike-east)))
        position))

(define (launch-trike! path)
  (set! enemies
    (aa-insert enemies
               (get-uid)
               (make <trike>
                 #:x (first (first path)) #:y (second (first path))
                 #:prev-x (first (first path)) #:prev-y (second (first path))
                 #:waypoints path))))

(define-class <tank> (<enemy>)
  (base-speed #:init-value 0.04 #:accessor base-speed)
  (max-health #:init-value 30 #:accessor .max-health))

(define-method (draw (e <tank>) position)
  (draw (get-sprite (case (.facing e)
                      ((#:west) #:tank-blue-t0-west)
                      ((#:north) #:tank-blue-t0-north)
                      ((#:south) #:tank-blue-t0-south)
                      ((#:east) #:tank-blue-t0-east)))
        position))

(define (launch-tank! path)
  (set! enemies
    (aa-insert enemies
               (get-uid)
               (make <tank>
                 #:x (first (first path)) #:y (second (first path))
                 #:prev-x (first (first path)) #:prev-y (second (first path))
                 #:waypoints path))))

(define-class <tank-l2> (<enemy>)
  (base-speed #:init-value 0.04 #:accessor base-speed)
  (max-health #:init-value 60 #:accessor .max-health))

(define-method (draw (e <tank-l2>) position)
  (draw (get-sprite (case (.facing e)
                      ((#:west) #:tank-yellow-t0-west)
                      ((#:north) #:tank-yellow-t0-north)
                      ((#:south) #:tank-yellow-t0-south)
                      ((#:east) #:tank-yellow-t0-east)))
        position))

(define (launch-tank-l2! path)
  (set! enemies
    (aa-insert enemies
               (get-uid)
               (make <tank-l2>
                 #:x (first (first path)) #:y (second (first path))
                 #:prev-x (first (first path)) #:prev-y (second (first path))
                 #:waypoints path))))

(define-class <tank-l3> (<enemy>)
  (base-speed #:init-value 0.04 #:accessor base-speed)
  (max-health #:init-value 90 #:accessor .max-health))

(define-method (draw (e <tank-l3>) position)
  (draw (get-sprite (case (.facing e)
                      ((#:west) #:tank-red-t0-west)
                      ((#:north) #:tank-red-t0-north)
                      ((#:south) #:tank-red-t0-south)
                      ((#:east) #:tank-red-t0-east)))
        position))

(define (launch-tank-l3! path)
  (set! enemies
    (aa-insert enemies
               (get-uid)
               (make <tank-l3>
                 #:x (first (first path)) #:y (second (first path))
                 #:prev-x (first (first path)) #:prev-y (second (first path))
                 #:waypoints path))))

(define-class <ptboat> (<enemy>)
  (base-speed #:init-value 0.1 #:accessor base-speed)
  (max-health #:init-value 10 #:accessor .max-health))

(define-method (draw (e <ptboat>) position)
  (draw (get-sprite (case (.facing e)
                      ((#:west) #:ptboat-west)
                      ((#:north) #:ptboat-north)
                      ((#:south) #:ptboat-south)
                      ((#:east) #:ptboat-east)))
        position))

(define (launch-ptboat! path)
  (set! enemies
    (aa-insert enemies
               (get-uid)
               (make <ptboat>
                 #:x (first (first path)) #:y (second (first path))
                 #:prev-x (first (first path)) #:prev-y (second (first path))
                 #:waypoints path))))

(define-class <destroyer> (<enemy>)
  (base-speed #:init-value 0.02 #:accessor base-speed)
  (max-health #:init-value 100 #:accessor .max-health))

(define-method (draw (e <destroyer>) position)
  (draw (get-sprite (case (.facing e)
                      ((#:west) #:destroyer-west)
                      ((#:north) #:destroyer-north)
                      ((#:south) #:destroyer-south)
                      ((#:east) #:destroyer-east)))
        position))

(define (launch-destroyer! path)
  (set! enemies
    (aa-insert enemies
               (get-uid)
               (make <destroyer>
                 #:x (first (first path)) #:y (second (first path))
                 #:prev-x (first (first path)) #:prev-y (second (first path))
                 #:waypoints path))))

(define-class <tracked-soldier> (<enemy>)
  (base-speed #:init-value 0.04 #:accessor base-speed)
  (max-health #:init-value 20 #:accessor .max-health))

(define-method (draw (e <tracked-soldier>) position)
  (draw (get-sprite (case (.facing e)
                      ((#:west) #:tracked-soldier-west)
                      ((#:north) #:tracked-soldier-north)
                      ((#:south) #:tracked-soldier-south)
                      ((#:east) #:tracked-soldier-east)))
        position))

(define (launch-tracked-soldier! path)
  (set! enemies
    (aa-insert enemies
               (get-uid)
               (make <tracked-soldier>
                 #:x (first (first path)) #:y (second (first path))
                 #:prev-x (first (first path)) #:prev-y (second (first path))
                 #:waypoints path))))

(define-class <helo> (<enemy>)
  (base-speed #:init-value 0.2 #:accessor base-speed)
  (max-health #:init-value 10 #:accessor .max-health))

(define-method (draw (e <helo>) position)
  (draw (get-sprite (case (.facing e)
                      ((#:west) #:helo-west)
                      ((#:north) #:helo-north)
                      ((#:south) #:helo-south)
                      ((#:east) #:helo-east)))
        position))

(define (launch-helo! path)
  (set! enemies
    (aa-insert enemies
               (get-uid)
               (make <helo>
                 #:x (first (first path)) #:y (second (first path))
                 #:prev-x (first (first path)) #:prev-y (second (first path))
                 #:waypoints path))))
